@@ -5,13 +5,15 @@ use crate::utils::{Client};
 
 use rayon::ThreadPool;
 use rayon::ThreadPoolBuilder;
+use rayon::iter::{IntoParallelRefIterator, ParallelExtend, ParallelIterator};
 use robotparser_fork::RobotFileParser;
 
-use std::collections::HashSet;
+use hashbrown::HashSet;
 use std::{sync, thread, time::Duration};
 use reqwest::header::CONNECTION;
 use reqwest::header;
 use sync::mpsc::{channel, Sender, Receiver};
+use log::{log_enabled, info, Level};
 
 /// Represent a website to scrawl. To start crawling, instanciate a new `struct` using
 /// <pre>
@@ -44,7 +46,7 @@ pub struct Website<'a> {
     pub page_store_ignore: bool,
 }
 
-type Message = (Page, Vec<String>);
+type Message = (Page, HashSet<String>);
 
 impl<'a> Website<'a> {
     /// Initialize Website object with a start link to scrawl.
@@ -66,6 +68,11 @@ impl<'a> Website<'a> {
         &self.pages
     }
 
+    /// links visited getter
+    pub fn get_links(&self) -> &HashSet<String> {
+        &self.links_visited
+    }
+
     /// crawl delay getter
     pub fn get_delay(&self) -> Duration {
         Duration::from_millis(self.configuration.delay)
@@ -119,17 +126,17 @@ impl<'a> Website<'a> {
                 if !self.is_allowed(link) {
                     continue;
                 }
-                self.log(&format!("- fetch {}", &link));
-                self.links_visited.insert(String::from(link));
+                if self.configuration.delay > 0 {
+                    thread::sleep(self.get_delay());
+                }
+                log("- fetch {}", link);
+
+                self.links_visited.insert(link.into());
 
                 let link = link.clone();
                 let tx = tx.clone();
                 let cx = client.clone();
 
-                if self.configuration.delay > 0 {
-                    thread::sleep(self.get_delay());
-                }
-                
                 pool.spawn(move || {
                     let link_result = on_link_find_callback(link);
                     let mut page = Page::new(&link_result, &cx);
@@ -145,17 +152,62 @@ impl<'a> Website<'a> {
 
             rx.into_iter().for_each(|page| {
                 let (page, links) = page;
-                self.log(&format!("- parse {}", page.get_url()));
-
+                log("- parse {}", page.get_url());
                 new_links.extend(links);
-
                 if !self.page_store_ignore {
                     self.pages.push(page);
                 }
-
             });
 
-            self.links = new_links.difference(&self.links_visited).cloned().collect();
+            self.links = &new_links - &self.links_visited;
+        }
+    }
+
+
+    /// Start to crawl website using the stack buffer [Experimental: may perform worse or better depending on system specs]
+    /// ATM API only modifies self.links_visited for extracting via website.get_links();
+    pub fn crawl_stack(&mut self, client: Option<Client>) {
+        self.configure_robots_parser();
+        let client = client.unwrap_or(self.configure_http_client(None));
+        let on_link_find_callback = self.on_link_find_callback;
+        
+        let (links_visited, new_links): (Vec<_>, Vec<_>) = self.links.par_iter().map(|link| {
+            let link_clone = link.clone();
+            let link_result = on_link_find_callback(link_clone);
+            let mut page = Page::new(&link_result, &client);
+            let page_links = page.links();
+            
+            (page.get_url().to_string(), page_links)
+        }).collect();
+
+        self.links_visited.par_extend(links_visited);
+
+        let current_links = &self.links_visited;
+        let black_list = &self.configuration.blacklist_url;
+        
+        // par filter and flatten new links
+        let new_links: HashSet<String> = new_links.par_iter().flatten().filter(|l| {
+            let link = l.to_string();
+            if current_links.contains(&link) {
+                return false
+            }
+            if contains(&black_list, &link) {
+                return false;
+            }
+            
+            return true
+        }).cloned().collect();
+
+        self.links = if self.configuration.respect_robots_txt {
+            // robots needs to be non static lifetime to allow parr usage.
+            new_links.iter().filter(|link| self.is_allowed_robots(&link)).cloned().collect()
+        } else {
+            new_links
+        };
+
+        // re-crawl if links exist
+        if !self.links.is_empty() {
+            self.crawl_stack(Some(client));
         }
     }
 
@@ -171,18 +223,19 @@ impl<'a> Website<'a> {
         if contains(&self.configuration.blacklist_url, link) {
             return false;
         }
-        if self.configuration.respect_robots_txt && !self.robot_file_parser.can_fetch("*", link) {
+        if self.configuration.respect_robots_txt && !self.is_allowed_robots(link) {
             return false;
         }
 
         true
     }
 
-    /// log to console if configuration verbose
-    pub fn log(&self, message: &str) {
-        if self.configuration.verbose {
-            println!("{}", message);
-        }
+
+    /// return `true` if URL:
+    ///
+    /// - is not forbidden in robot.txt file (if parameter is defined)  
+    pub fn is_allowed_robots(&self, link: &String) -> bool {
+        self.robot_file_parser.can_fetch("*", link)
     }
 }
 
@@ -190,6 +243,13 @@ impl<'a> Drop for Website<'a> {
     fn drop(&mut self) {}
 }
 
+/// log to console if configuration verbose
+pub fn log(message: &str, data: impl AsRef<str>) {
+    if log_enabled!(Level::Info) {
+        info!("{}{}", &message, data.as_ref());
+    }
+}
+
 #[test]
 fn crawl() {
     let mut website: Website = Website::new("https://choosealicense.com");
@@ -203,6 +263,19 @@ fn crawl() {
     );
 }
 
+#[test]
+fn crawl_stack() {
+    let mut website: Website = Website::new("https://choosealicense.com");
+    website.crawl_stack(None);
+    assert!(
+        website
+            .links_visited
+            .contains(&"https://choosealicense.com/licenses/".to_string()),
+        "{:?}",
+        website.links_visited
+    );
+}
+
 #[test]
 fn crawl_invalid() {
     let url = "https://w.com";
@@ -218,7 +291,7 @@ fn crawl_invalid() {
 fn crawl_link_callback() {
     let mut website: Website = Website::new("https://choosealicense.com");
     website.on_link_find_callback = |s| {
-        println!("callback link target: {}", s);
+       log("callback link target: {}", &s);
         s
     };
     website.crawl();
